#!/usr/bin/python3

# This file is part of Contacts daemon
#
# Copyright (c) 2011 Nokia Corporation and/or its subsidiary(-ies).
#
# Contact:  Nokia Corporation (info@qt.nokia.com)
#
# GNU Lesser General Public License Usage
# This file may be used under the terms of the GNU Lesser General Public License
# version 2.1 as published by the Free Software Foundation and appearing in the
# file LICENSE.LGPL included in the packaging of this file.  Please review the
# following information to ensure the GNU Lesser General Public License version
# 2.1 requirements will be met:
# http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
#
# In addition, as a special exception, Nokia gives you certain additional rights.
# These rights are described in the Nokia Qt LGPL Exception version 1.1, included
# in the file LGPL_EXCEPTION.txt in this package.
#
# Other Usage
# Alternatively, this file may be used in accordance with the terms and
# conditions contained in a signed written agreement between you and Nokia.

import sys

mode = 0

lines = sys.stdin.readlines()

for line in lines:

    if mode == 3:
       mode = 0
       continue

    if mode == 2:
        mode = 3
        continue

    if mode == 1:
        mode = 2
        if last.find('.moc') > 0:
            continue
        
        if last[-6:-2] != ".cpp":
            continue
        
        print(last[:-1], ":", line[:-1])
        continue

    if mode == 0:
        mode = 1
        last = line
        continue
